# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/Layers.ipynb.

# %% auto 0
__all__ = ['dummy']

# %% ../../../nbs/Layers.ipynb 1
from ..._components.mono_dense_layer import MonoDense

# %% ../../../nbs/Layers.ipynb 2
def dummy() -> None:
    pass


dummy.__module__ = "_dummy"

# %% ../../../nbs/Layers.ipynb 3
__all__ = ["MonoDense"]
